/**
* Copyright (c) 2021-2022 Kedron Holdings LLC, All Rights Reserved.
* This file is not part of Reactor, but a free add-on for it. This file only is offered under the MIT License.
* See https://mit-license.org for license text/details.
*/

const version = 22323;

const Configuration = require("server/lib/Configuration");
const Logger = require("server/lib/Logger");
const Controller = require("server/lib/Controller").requires( 22125 );
const { coalesce } = require( "server/lib/util" );

Logger.getLogger( 'OctoPrintController', Logger.getLogger( 'Controller' ) )
    .always( "Module OctoPrintController v%1", version );

const basedir = Configuration.getConfig( "reactor.basedir" );
const dumpdir = Configuration.getConfig( "reactor.logsdir" );
const confdir = Configuration.getConfig( "reactor.confdir" );

module.exports = class OctoPrintController extends Controller {
    constructor( struct, id, config ) {
        super( struct, id, config );

        this.num_fail = 0;
        this.control_entity = false;
        this.control_enabled = true;
        this.booting = false;

        /* Since this Controller only represents one thing, use the system entity to carry it. */
        this.removeControllerGroup();
        this.system.extendCapability( 'printer3d' );
    }

    /** Start the controller. */
    async start() {
        if ( this.config.control_entity ) {
            let e = this.getStructure().findEntity( this.config.control_entity );
            if ( !e ) {
                this.log.err( "%1 control entity %2 not found", this, this.config.control_entity );
                this.sendError( "The configured control entity {0:q} was not found",
                    this.config.control_entity );
            } else {
                this.control_entity = e;
                this.subscribe( e );
                this.control_enabled = e.getAttribute( 'power_switch.state' ) !== false;  /* null OK */
                this.log.info( "%1 control entity state: %2", this, this.control_enabled );
            }
        }
        this.run();
        this.online();  /* unconditionally */
        return this;
    }

    notify( event ) {
        try {
            this.log.debug( 6, "%1 event notification %2", this, event );
            if ( this.control_entity && "entity-changed" === event.type &&
                this.control_entity.getCanonicalID() === event.sender.getCanonicalID() ) {
                this.log.debug( 5, "%1 getting control entity %2 state", this, this.control_entity );
                const newstate = this.control_entity.getAttribute( 'power_switch.state' ) !== false;  /* null OK */
                if ( newstate !== this.control_enabled ) {
                    this.control_enabled = newstate;
                    this.log.debug( 5, "%1 control entity state is now %2", this, this.control_enabled );
                    if ( newstate ) {
                        this.booting = true;
                    }
                    this.run();
                }
            }
        } catch ( err ) {
            this.log.err( "%1 error handling event %2 from %3: %4", this, event.type, event.sender, err );
            this.log.exception( err );
        }

        super.notify( event );  /* MANDATORY */
    }

    run() {
        const e = this.system;
        if ( this.control_entity && ! this.control_enabled ) {
            this.log.debug( 5, "%1 printer control device %2 is off", this, this.control_entity );
            this._printer_offline( false, "Control power off" );
            this.num_fail = 0;
            this.booting = false;
            this.stopTimer();
            return;
        }

        this.log.debug( 5, "%1 querying printer at %2", this, this.config.source );
        this.fetchJSON( this.config.source + '/api/printer',
            {
                timeout: this.config.timeout || 15000,
                headers: { 'x-api-key': this.config.api_key }
            }
        ).then( data => {
            this.num_fail = 0;
            this.booting = false;
            this.log.debug( 5, "%1: received response data %2", this, data );
            e.deferNotifies( true );
            e.markDead( false );
            try {
                e.setAttribute( 'printer3d.reachable', true );
                e.setAttribute( 'printer3d.printing', coalesce( data.state?.flags?.printing ) );
                e.setAttribute( 'printer3d.paused', coalesce( data.state?.flags?.paused ) );
                e.setAttribute( 'printer3d.resuming', coalesce( data.state?.flags?.resuming ) );
                e.setAttribute( 'printer3d.ready', coalesce( data.state?.flags?.ready ) );
                e.setAttribute( 'printer3d.operational', coalesce( data.state?.flags?.operational ) );
                e.setAttribute( 'printer3d.error', coalesce( data.state?.flags?.error ) );
                e.setAttribute( 'printer3d.finishing', coalesce( data.state?.flags?.finishing ) );
                e.setAttribute( 'printer3d.temperature_bed_actual', coalesce( data.temperature?.bed?.actual ) );
                e.setAttribute( 'printer3d.temperature_bed_target', coalesce( data.temperature?.bed?.target ) );
                e.setAttribute( 'printer3d.temperature_hotend_actual', coalesce( data.temperature?.tool0?.actual ) );
                e.setAttribute( 'printer3d.temperature_hotend_target', coalesce( data.temperature?.tool0?.target ) );
                e.setAttribute( 'printer3d.message', coalesce( data.state?.text ) );
                e.setAttribute( 'printer3d.sdcard_ready', coalesce( data.sd?.ready ) );
            } catch ( err ) {
                this.log.exception( err );
            } finally {
                e.deferNotifies( false );
                this.startDelay( this.config.interval || 60000 );
            }
        }).catch( err => {
            this.log.err( "%1: error fetching printer state: %2", this, err );
            if ( this.booting ) {
                this.startDelay( this.config.boot_interval || 30000 );
                this._printer_offline( false, "Control power on; waiting for printer ready" );
            } else {
                this.startDelay( Math.min( 120000, ( this.config.error_interval || 5000 ) * Math.max( 1, ++this.num_fail - 12 ) ) );
                if ( this.num_fail >= 3 ) {
                    this._printer_offline( true, `${this.config.source} ${String(err)}` );
                }
            }
        });
    }

    _printer_offline( isError, msg ) {
        const e = this.system;
        this.log.debug( 5, "%1 setting printer offline, error=%2: %3", this, !! isError, msg );
        e.setAttributes({
            reachable: false,
            printing: false,
            paused: false,
            resuming: false,
            ready: false,
            operational: false,
            error: !! isError,
            finishing: false,
            temperature_bed_actual: null,
            temperature_bed_target: null,
            temperature_hotend_actual:  null,
            temperature_hotend_target: null,
            message: msg || null,
            sdcard_ready: null
        }, 'printer3d' );
        e.deferNotifies( false );
    }
};
